# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

import fnmatch
import inspect
import itertools
import os
import re
import sys
import traceback
from pydoc import locate
import glob
import pkgutil

from htf._decorators import _is_cython
from htf.termcolors import system
from htf.core import Test

PY_3 = sys.version_info >= (3,)

if PY_3:  # pragma: no cover
    import importlib.util

    def load_module(file):
        module_name = os.path.splitext(os.path.basename(file))[0]
        spec = importlib.util.spec_from_file_location(module_name, file)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        try:
            spec.loader.exec_module(module)
        except Exception as exc:
            exc.__cause__ = None
            raise DiscoveryError('Loading module "{}" failed with the following error:\n\n'
                                 '{}'.format(file, ''.join(traceback.format_exception(*sys.exc_info()))))
        return module
else:  # pragma: no cover
    import imp

    def load_module(file):
        try:
            module = imp.load_source(os.path.splitext(os.path.basename(file))[0], file)
        except Exception as exc:
            exc.__cause__ = None
            raise DiscoveryError('Loading module "{}" failed with the following error:\n\n'
                                 '{}'.format(file, ''.join(traceback.format_exception(*sys.exc_info()))))
        return module


class DiscoveryError(Exception):
    def __init__(self, message):
        super(DiscoveryError, self).__init__(message)
        self.__cause__ = None  # disable exception chaining in Python 3


class Discovery(object):
    """
    A class responsible for resolving given specifiers into corresponding test and fixture objects.
    """
    def __init__(self, test_specifiers, fixture_specifiers, file_pattern='test*.py'):
        """
        Args:
            test_specifiers(list of str): strings specifying tests or test locations
            fixture_specifiers(list of str):  strings specifying fixtures or fixture locations
            file_pattern (str): pattern to use to discover files in given folders
        """
        self._test_specifiers = test_specifiers
        self._fixture_specifiers = fixture_specifiers
        self._file_pattern = file_pattern

    @staticmethod
    def _is_test(obj):
        """
        Checks if a given object should be considered a test.

        Args:
            obj (object): object to inspect

        Returns:
            True if the object is likely a test, False otherwise
        """
        return (inspect.isfunction(obj) or inspect.ismethod(obj) or _is_cython(obj)) \
            and (obj.__name__.startswith('test') or hasattr(obj, "htf_test")) \
            and not hasattr(obj, "htf_fixture")

    @staticmethod
    def _get_tests_from_object(obj):
        """
        Returns a list of test functions for a given object. A test function is any function or method whose name
        starts with 'test'.

        Args:
            obj (object): object to look for tests in

        Returns:
            List of test functions
        """

        return [Test(test, obj) for test in [getattr(obj, attr) for attr in dir(obj)] if Discovery._is_test(test)]

    def _get_tests_from_module(self, module):
        """
        Return a list of test functions for a given module.

        Args:
            module: module to look for tests in

        Returns:
            List of test functions
        """
        tests = []
        if hasattr(module, '__path__'):
            for mod in pkgutil.walk_packages(module.__path__, prefix=module.__name__+'.'):
                module_pattern = self._file_pattern.replace('.py', '').replace('*', '.*')
                if re.match(module_pattern, mod[1].split('.')[-1]):
                    try:
                        imported = __import__(mod[1], fromlist='dummy')
                    except Exception as exc:
                        exc.__cause__ = None
                        raise DiscoveryError('Importing test module "{}" failed with the following error:\n\n'
                                             '{}'.format(module, ''.join(traceback.format_exception(*sys.exc_info()))))
                    tests += self._get_tests_from_module(imported)
        tests += list(itertools.chain(*map(self._get_tests_from_object,
                                           [val for name, val in inspect.getmembers(module, predicate=inspect.isclass)])))
        tests += self._get_tests_from_object(module)
        return tests

    @staticmethod
    def _get_fixtures_from_object(obj):
        """
        Return a list of fixtures for a given object. A fixture is any function or class
        with the 'htf_fixture' attribute.

        Args:
            obj: object to look for fixtures in

        Returns:
            List of fixtures
        """
        def _is_fixture(attr):
            return (inspect.isfunction(attr) or inspect.isclass(attr)) \
                   and hasattr(obj, 'htf_fixture') \
                   and not hasattr(obj, "htf_test")

        return [obj] if _is_fixture(obj) else []

    def _get_fixtures_from_module(self, module):
        """
        Return a list of fixtures for a given module.

        Args:
            module: module to look for fixtures in

        Returns:
            List of fixtures
        """
        fixtures = []
        if hasattr(module, '__path__'):
            for mod in pkgutil.walk_packages(module.__path__, prefix=module.__name__+'.'):
                try:
                    imported = __import__(mod[1], fromlist=str('dummy'))
                except Exception as exc:
                    exc.__cause__ = None
                    raise DiscoveryError('Importing fixture module "{}" failed with the following error:\n\n'
                                         '{}'.format(module, ''.join(traceback.format_exception(*sys.exc_info()))))
                fixtures += self._get_fixtures_from_module(imported)
        fixtures += list(itertools.chain(*map(self._get_fixtures_from_object,
                                              [val for name, val in inspect.getmembers(module, predicate=inspect.isclass)])))
        fixtures += list(itertools.chain(*map(self._get_fixtures_from_object,
                                              [val for name, val in inspect.getmembers(module, predicate=inspect.isfunction)])))
        return fixtures

    @staticmethod
    def _get_modules_from_file(file):
        """
        Return a list of modules given a file specifier. The file specifier can contain wildcards.

        Args:
            file (str): file specifier

        Returns:
            List of module objects
        """
        modules = []
        path = os.path.abspath(file)
        files = glob.glob(path)
        for file in files:
            modules.append(load_module(file))
        return modules

    def _resolve_specifiers(self, specifiers, do_print=True, filter_files=False):
        """
        Resolves specifiers into modules, classes or functions depending on the specifier.

        Args:
            specifiers(list of str): specifiers to resolve

        Returns:
            List of modules, classes and functions
        """
        resolved_specifiers = []
        for specifier in specifiers:
            if do_print is True:
                print(system(specifier))

            if isinstance(specifier, (type(""), bytes)):
                try:
                    if os.path.isdir(specifier):
                        sys.path.append(specifier)
                        resolved_specifiers += self._get_modules_from_directory(specifier, filter_files)
                        continue
                except TypeError:  # pragma: no cover
                    pass

                try:
                    obj = locate(specifier)
                except Exception as exc:
                    exc.__cause__ = None  # disable exception chaining of the pydoc error
                    raise DiscoveryError('Locating specifier "{}" failed with the following error:\n\n'
                                         '{}'.format(specifier, ''.join(traceback.format_exception(*sys.exc_info()))))
                if obj is not None:
                    resolved_specifiers.append(obj)
                    continue

                try:
                    if os.path.isfile(specifier) or "*" in specifier:
                        resolved_specifiers += self._get_modules_from_file(specifier)
                        continue
                except TypeError:  # pragma: no cover
                    pass

                raise DiscoveryError('Could not resolve specifier: {}'.format(specifier))

            resolved_specifiers.append(specifier)
        return resolved_specifiers

    def _get_modules_from_directory(self, directory, filter_files):
        """
        Return a list of modules for a given directory.

        Args:
            directory (str): path to the directory

        Returns:
            List of module objects
        """
        modules = []
        path = os.path.abspath(directory)
        files = []
        for root, dirnames, filenames in os.walk(path):
            if filter_files:
                for filename in fnmatch.filter(filenames, self._file_pattern):
                    files.append(os.path.join(root, filename))
            else:
                for filename in fnmatch.filter(filenames, '*.py'):
                    files.append(os.path.join(root, filename))

        for file in files:
            modules.append(load_module(file))
        return modules

    @staticmethod
    def _add_cwd_to_path():
        cwd = os.getcwd()
        if cwd not in sys.path:
            sys.path.insert(0, cwd)

    def discover_tests(self):
        """
        Discover test functions using the test specifiers provided.

        Returns:
            List of test functions
        """

        self._add_cwd_to_path()
        tests = []

        for specifier in self._resolve_specifiers(self._test_specifiers, filter_files=True):
            if inspect.ismodule(specifier):
                tests += self._get_tests_from_module(specifier)

            elif self._is_test(specifier):
                tests.append(Test(specifier))

            else:
                tests += self._get_tests_from_object(specifier)
        return tests

    def discover_fixtures(self):
        """
        Discover fixtures using the fixture specifiers provided.

        Returns:
            List of fixtures
        """
        self._add_cwd_to_path()
        fixtures = []

        for specifier in self._resolve_specifiers(self._fixture_specifiers, do_print=False):
            if inspect.ismodule(specifier):
                fixtures += list(set(self._get_fixtures_from_module(specifier)))

            elif inspect.isclass(specifier) or inspect.isfunction(specifier):
                fixtures += self._get_fixtures_from_object(specifier)

        return fixtures

